/*
    ArchiveSMTP Milter v1.0.b1
    --------------------------
	Copyright (c) 2008 Dancing Fortune Software. All rights reserved.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions
	are met:
 
	1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
 
	2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
	POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <time.h>
#include <unistd.h>

#include <libmilter/mfapi.h>

/* Does this system define a type for bool? */
#ifndef bool
#define bool	int
#define TRUE	1
#define FALSE	0
#endif

enum
{
	kFieldUndefined,
	kFieldSMTPSender,
	kFieldFromAddress,
	kFieldToAddress
};

enum
{
	kLocationFieldUndefined,
	kLocationFieldStart,
	kLocationFieldEnd,
	kLocationFieldAny
};

/*
 Test
 
 The test class describes a test to be performed on a field. There are three elements to a test.
 
 1. the field to be tested;
 - SMTP sender
 - from address
 - to address
 2. the location in the field to be tested;
 - start of field
 - end of field
 - anywhere in field
 3. the text to be searched for.
 
 This test can then be performed with the PerformTest() function, which returns a true or false
 based on whether the text succeeded or failed.
 */
typedef struct Test_s
{
	int		fieldToTest;
	int		locationToTest;
	char*	textToFind;
} Test;

typedef struct Condition_s Condition;
/*
	Condition_s
	
	This lives above rules on the tree
	
		 Test   Test
			\    /
	Test  Condition
	 \    /
	Condition
		|
		|
	  Rule
	
	A Condition describes an OR or AND performed on a combination of tests, other conditions
	or nothing. It consists of two pointers, each to either a test or a condition. These are
	evaluated to determine the status of the condition.
	
	The function in question is EvaluateCondition(Condition*)
	
	A Condition can be initialised with InitialiseCondition(Condition*)
 */
struct Condition_s
{
	int			type;
	Condition	*p1, *p2;
	Test		*t1, *t2; /* Quirk: the initial design called for it, but
							  in practice *t2 isn't ever actually used. */
};

enum
{
	kConditionTypeUndefined, /*  condition type is not yet defined and
								testing it will produce an error	  */
	kConditionTypeOr,		 /*  condition is an OR of two other
								conditions or tests					  */
	kConditionTypeAnd		 /*  condition is an AND of two other
								conditions or tests					  */
};

typedef struct Rule_s Rule;
struct Rule_s
{
	Condition	*root;
	char		*storeIn;
	Rule		*next;
};

/* Convinence Macro to retrive private data instance */
#define MLFIPRIV	((mlfiPriv_t *) smfi_getpriv(ctx))

/* Verbose level checking and printing Macro */
#define VerbosePrint(L, STR, ARG) if (verboseLevel >= L) { printf(STR, ARG); }

/* Global Variables */
Rule		*firstRule;			/* Start of rule list */
int			verboseLevel;		/* Verbosity level */
bool		alwaysAccept;		/* Accept policy (default to reject on error) */

/* Mail To address list */
typedef struct toAddrList_s
{
	char	*address;			/* The address value */
	struct	toAddrList_s *next;	/* Link to next address */
} toAddrList_t;

/* Mail header structure */
typedef struct headerList_s
{
    char    *headerName;		/* The header field name */
    char    *headerValue;		/* The value of the field */
    struct  headerList_s *next;	/* Link to the next header field */
} headerList_t;

/* Mail body data structure */
typedef struct bodyChunks_s
{
    char    *body;				/* Pointer to body chunk data */
    int     size;				/* Size in bytes of body data */
    struct  bodyChunks_s *next;	/* Link to the next chunk of body data */
} bodyChunks_t;

/* Private instance data */
typedef struct mlfiPriv
{
	toAddrList_t	*mlfi_toAddr_list;	/* Linked list of recipient addresses */
	headerList_t    *mlfi_header_list;	/* Linked list of mail headers */
	bodyChunks_t    *mlfi_body_chunks;	/* Linked list of mail body data */
	char            *mlfi_username;		/* User name of mail sender */
    char            *mlfi_useraddr;		/* From address of mail sender */
    int             mlfi_bytes;			/* Total size of mail in bytes */
} mlfiPriv_t;


/*** Function prototypes ***/

/* Milter call-backs */
sfsistat mlfi_envfrom	(SMFICTX*, char**);
sfsistat mlfi_envrcpt	(SMFICTX*, char**);
sfsistat mlfi_header	(SMFICTX*, char*, char*);
sfsistat mlfi_body		(SMFICTX*, unsigned char*, size_t);
sfsistat mlfi_eom		(SMFICTX*);
sfsistat mlfi_abort		(SMFICTX*);
sfsistat mlfi_cleanup	(SMFICTX*, bool);

/* Configuration reading & Rule building/deleting */
void ReadConfigurationFile	(const char* filename);
void InitialParsingState	(FILE* file);
void RuleStartedParsingState(FILE* file, Rule* theRule);
Rule* CreateNewRule			();
void AddRuleToRuleList		(Rule* newRule);
void DeleteRule				(Rule* r);
Condition* ParseCondition	(FILE* file);
void InitialiseCondition	(Condition* cond);
void DeleteCondition		(Condition* c);
Test* ParseTest				(FILE* file, int testField);

/* Rule testing & printing */
int EvaluateRule			(Rule* r, mlfiPriv_t *priv);
int EvaluateCondition		(Condition* cond, mlfiPriv_t *priv);
int EvaluateConditionParameter(Condition* c, Test* t, mlfiPriv_t *priv);
int PerformTest				(Test* t, const char* testData);
void PrintCondition			(Condition* c);
void PrintTest				(Test* t);

/* Utility */
char* ParseMailAddr			(const char *addr);
void MemErrorPrint			(const char *func, const char *obj, int type);
int CheckStringIsNextInFile	(FILE* file, const char* string, const char* errorMessage);
char EatWhitespace			(FILE* file);
char* EatWhiteAndReadString	(FILE* file);
void StringToLower			(char *str);
static void usage			(char *prog);


/* mlfi_envfrom:
 Called on "mail from" cmd from client, sets up private data block for
 this mail session and determines local user name of sender.
 */
sfsistat mlfi_envfrom(SMFICTX *ctx, char **argv)
{
    /* Set up our private data block for this session */
    mlfiPriv_t *priv = malloc(sizeof(mlfiPriv_t));
    
    /* Did we allocate memory? */
    if (priv == NULL) {
        MemErrorPrint("mlfi_envfrom", "private data", 0);
		if (alwaysAccept)
			return SMFIS_CONTINUE;
		else
			return SMFIS_TEMPFAIL; /* Reject mail with temporary error */
    }
    
    /* Set initial pointers to NULL */
	priv->mlfi_toAddr_list = NULL;
    priv->mlfi_header_list = NULL;
    priv->mlfi_body_chunks = NULL;
    priv->mlfi_username = NULL;
    priv->mlfi_useraddr = NULL;
    priv->mlfi_bytes = 0;
    
    /* Associate data block with this messages context */
	smfi_setpriv(ctx, priv);
    
    /* Get the username of this mail client and copy to our struct */
    char *username;
    
    username = smfi_getsymval(ctx, "{auth_authen}");
    if (username != NULL) {
		/* Username was set, store it */
		char *usercopy = malloc(sizeof(char)*(strlen(username)+1));
		strcpy(usercopy, username);
		priv->mlfi_username = usercopy;
	}
	
	/* Store from address */
	char *addrcopy = ParseMailAddr(argv[0]);
	if (addrcopy == NULL) {
		/* Reject mail with temporary failure if we can't allocate memory */
		MemErrorPrint("mlfi_envfrom", "sender address", 0);
		return mlfi_cleanup(ctx, FALSE);
	}
    priv->mlfi_useraddr = addrcopy;
    
    /* To help with file locking later, calculate the size of the message so far,
	 it should start with the from address plus an extra 9 characters for the word
	 "from" then two spaces followed by the opening < and closing > */
	priv->mlfi_bytes += ((strlen(addrcopy) + 9) * sizeof(char));
    
	/* Continue processing */
	return SMFIS_CONTINUE;
}

/* mlfi_envrcpt:
 Called on "rcpt to" cmd from client. We will collect this data later with the
 rest of the headers, only here for debugging purposes.
 */
sfsistat mlfi_envrcpt(SMFICTX *ctx, char **argv)
{
    /* Grab our private session data */
    mlfiPriv_t *priv = MLFIPRIV;
    if (priv == NULL) {
		/* This shouldn't ever happen, but to be safe we test for it */
        MemErrorPrint("mlfi_envrcpt", NULL, 1);
        if (alwaysAccept)
			return SMFIS_CONTINUE;
		else
			return SMFIS_TEMPFAIL;
    }
	
	/* Add the recipient address to the list */
	toAddrList_t *address = NULL;
	if (argv[0] != NULL) {
		address = malloc(sizeof(toAddrList_t));
		if (address == NULL) {
			/* Reject mail with temporary failure if we can't allocate memory */
			MemErrorPrint("mlfi_envrcpt", "recipient structure", 0);
			return mlfi_cleanup(ctx, FALSE);
		}
		
		address->address = ParseMailAddr(argv[0]);
		if (address->address == NULL) {
			/* Reject mail with temporary failure if we can't allocate memory */
			MemErrorPrint("mlfi_envrcpt", "recipient address", 0);
			return mlfi_cleanup(ctx, FALSE);
		}
		address->next = NULL;
	}
	
	/* Add this address struct to list */
    if (priv->mlfi_toAddr_list == NULL) {
        priv->mlfi_toAddr_list = address;
	}
    else {
        /* Iterate through list until we find the last item */
        toAddrList_t *ptr = priv->mlfi_toAddr_list;
        while (ptr->next != NULL) {
            ptr = ptr->next;
		}
        ptr->next = address;
    }
	
    return SMFIS_CONTINUE;
}

/* mlfi_header:
 Called for each header from client, copies the headers into our private
 data block as a linked list with header field name and value. Note: the
 trailing CR/LF (\r\n) is removed from the value.
 */
sfsistat mlfi_header(SMFICTX *ctx, char *headerf, char *headerv)
{
    /* Grab our private session data */
    mlfiPriv_t *priv = MLFIPRIV;
    if (priv == NULL) {
		/* This shouldn't ever happen, but to be safe we test for it */
		MemErrorPrint("mlfi_header", NULL, 1);
		if (alwaysAccept)
			return SMFIS_CONTINUE;
		else
			return SMFIS_TEMPFAIL;
	}
    
    /* Allocate and fill header stucture memory */
    if (headerf == NULL) {
		/* Is this possible? Check for it just in case outlook decides it is */
        fprintf(stderr, "ArchiveSMTP(mlfi_header): WARNING - Header name is not set, ignoring this header.\n");
        return SMFIS_CONTINUE;
    }
    
	headerList_t *header = malloc(sizeof(headerList_t));
    header->headerName = malloc(sizeof(char)*(strlen(headerf)+1));
    if (header == NULL || header->headerName == NULL) {
		/* Reject mail with temporary failure if we can't allocate memory */
		MemErrorPrint("mlfi_header", "header name", 0);
        return mlfi_cleanup(ctx, FALSE);
    }
	
    strcpy(header->headerName, headerf);
	
	/* Check there is a header value (outlook sometimes puts headers with no value),
	 if there is a value allocate memory and store it. */
    if (headerv != NULL) {
        header->headerValue = malloc(sizeof(char)*(strlen(headerv)+1));
        if (header->headerValue == NULL) {
            /* Reject mail with temporary failure if we can't allocate memory */
			MemErrorPrint("mlfi_header", "header value", 0);
			return mlfi_cleanup(ctx, FALSE);
        }
        strcpy(header->headerValue, headerv);
    }
	
    header->next = NULL;
    
    /* Add this header struct to list */
    if (priv->mlfi_header_list == NULL) {
        priv->mlfi_header_list = header;
	}
    else {
        /* Iterate through list until we find the last item */
        headerList_t *ptr = priv->mlfi_header_list;
        while (ptr->next != NULL) {
            ptr = ptr->next;
		}
        ptr->next = header;
    }
    
    /* Add field sizes to total message size (plus 3 for ": " and \n) */
    priv->mlfi_bytes += ((strlen(header->headerName) + strlen(header->headerValue)
                         + 3) * sizeof(char));
    
	/* Continue processing */
	return SMFIS_CONTINUE;
}

/* mlfi_body:
 Called for each body chunk from client, can be called multiple times and
 data is not c string formatted. Copies body chunks into private data struct
 as a linked list. 
 */
sfsistat mlfi_body(SMFICTX *ctx, unsigned char *bodyp, size_t bodylen)
{
    /* Grab our private session data */
    mlfiPriv_t *priv = MLFIPRIV;
    if (priv == NULL) {
		/* This shouldn't ever happen, but to be safe we test for it */
		MemErrorPrint("mlfi_body", NULL, 1);
		if (alwaysAccept)
			return SMFIS_CONTINUE;
		else
			return SMFIS_TEMPFAIL;
	}
    
    /* Allocate and fill struct for this body chunk */
    bodyChunks_t *body = malloc(sizeof(bodyChunks_t));
    if (body == NULL) {
        /* Reject mail with temporary failure if we can't allocate memory */
		MemErrorPrint("mlfi_body", "body structure", 0);
        return mlfi_cleanup(ctx, FALSE);
    }
    body->body = malloc(bodylen);
    if (body->body == NULL) {
        /* Reject mail with temporary failure if we can't allocate memory */
		MemErrorPrint("mlfi_body", "body data", 0);
        return mlfi_cleanup(ctx, FALSE);
    }
	
	/* Body is NOT a string and can contain \0, so we must use memcpy */
    memcpy(body->body, bodyp, bodylen);
    body->size = bodylen;
    body->next = NULL;
    
    /* Add this body chunk to the list */
    if (priv->mlfi_body_chunks == NULL) {
        priv->mlfi_body_chunks = body;
	}
    else {
        /* Iterate through list to find last item */
        bodyChunks_t *ptr = priv->mlfi_body_chunks;
        while (ptr->next != NULL) {
            ptr = ptr->next;
		}
        ptr->next = body;
    }
    
    /* Add body size to total size */
    priv->mlfi_bytes += bodylen;
    
	/* Continue processing */
	return SMFIS_CONTINUE;
}

/* mlfi_eom:
 Called at the end of a message from client, at this point we should have got
 all the message data, so we can write it to the correct user owned archive.
 */
sfsistat mlfi_eom(SMFICTX *ctx)
{
    /* Grab our private session data */
    mlfiPriv_t *priv = MLFIPRIV;
    if (priv == NULL) {
		/* This shouldn't ever happen, but to be safe we test for it */
		MemErrorPrint("mlfi_eom", NULL, 1);
		if (alwaysAccept)
			return SMFIS_CONTINUE;
		else
			return SMFIS_TEMPFAIL;
	}
	
	/* Now we must evaluate the collected data against our rule set to see if
	 and where to store this e-mail. */
	Rule *itr = firstRule;
	while (itr) 
	{
		if (EvaluateRule(itr, priv))
		{
			/* Generate timestamp */
			time_t stamp = time(NULL);
			/* Add size of time stamp to total size (plus 3 for eom padding and blank line after headers) */
			priv->mlfi_bytes += ((strlen(ctime(&stamp)) + 3) * sizeof(char));
			
			/* Open the correct file */
			char *userpath = itr->storeIn;
			
			/* Open File Descriptor */
			int outFD = open(userpath, O_WRONLY|O_CREAT, S_IRUSR|S_IWUSR|S_IWGRP);
			if (outFD == -1) {
				/* We couldn't open the file, so call the clean up routine */
				char msg[strlen(userpath)+24];
				sprintf(msg, "Could not open file '%s'.", userpath);
				MemErrorPrint("mlfi_eom", msg, 2);
				return mlfi_cleanup(ctx, FALSE);
			}
			
			/* Obtain lock on current EOF to num bytes past that we need to write */
			struct flock opts;
			opts.l_start    = 0;        /* starting offset */
			opts.l_len      = priv->mlfi_bytes; /* len = 0 means until end of file */
			opts.l_type     = F_WRLCK;  /* lock type: read/write, etc. */
			opts.l_whence   = SEEK_END; /* type of l_start */
			
			/* Keep trying to get lock while error val is: locked (EAGAIN) */
			while (fcntl(outFD, F_SETLK, &opts) == -1) {
				if (errno != EAGAIN) {
					/* There was a dead lock or some other error, call clean up */
					char msg[strlen(userpath)+32];
					sprintf(msg, "Could not get write lock on '%s'.", userpath);
					MemErrorPrint("mlfi_eom", msg, 2);
					return mlfi_cleanup(ctx, FALSE);
				}
			}
			
			/* Connect stream to FD */
			FILE *output = fdopen(outFD, "a");
			if (output == NULL) {
				/* We failed to connect the stream, so call clean up */
				MemErrorPrint("mlfi_eom", "Could not connect stream to FD.", 2);
				return mlfi_cleanup(ctx, FALSE);
			}
			
			/* Write mbox header */
			fwrite("From ", sizeof(char), strlen("From "), output);
			if (priv->mlfi_useraddr == NULL) {
				/* We might want to do something about this, but currently ignore*/
			} else {
				fwrite("<", sizeof(char), strlen("<"), output);
				fwrite(priv->mlfi_useraddr, sizeof(char), strlen(priv->mlfi_useraddr), output);
				fwrite(">", sizeof(char), strlen(">"), output);
			}
			fwrite("  ", sizeof(char), strlen("  "), output);
			fwrite(ctime(&stamp), sizeof(char), strlen(ctime(&stamp)), output);
			
			/* Write Headers to File */
			headerList_t *ptrH = priv->mlfi_header_list;
			if (ptrH != NULL) {
				do {
					fwrite(ptrH->headerName, sizeof(char), strlen(ptrH->headerName), output);
					fwrite(": ", sizeof(char), strlen(": "), output);
					fwrite(ptrH->headerValue, sizeof(char), strlen(ptrH->headerValue), output);
					fwrite("\n", sizeof(char), strlen("\n"), output);
					ptrH = ptrH->next;
				} while (ptrH != NULL);
			}
			else {
				/* Having no headers should be impossible */
				fprintf(stderr, "ArchiveSMTP(mlfi_eom): WARNING - Message contains no headers.\n");
			}
			
			/* Blank line after headers */
			fwrite("\n", sizeof(char), strlen("\n"), output);
			
			/* Write Body to File */
			bodyChunks_t *ptrB = priv->mlfi_body_chunks;
			if (ptrB != NULL) {
				do {
					fwrite(ptrB->body, sizeof(char), ptrB->size, output);
					ptrB = ptrB->next;
				} while (ptrB != NULL);
			}
			
			/* Blank line after body */
			fwrite("\n\n", sizeof(char), strlen("\n\n"), output);
			
			/* Unlock file area */
			if (fcntl(outFD, F_UNLCK, &opts) == -1) {
				fprintf(stderr, "ArchiveSMTP(mlfi_eom): WARNING - Could not release write lock. (errno: %d:%s)\n", errno, strerror(errno));
			}
			
			/* Close stream (will also close fd) */
			fclose(output);
		}
		itr = itr->next;
		VerbosePrint(1, "%s", "\n");
	}
    
    /* Call clean-up function */
	return mlfi_cleanup(ctx, TRUE);
}

/* mlfi_abort:
 Called when for what ever reason a message is aborted before being sent, we
 will just discard any aborted mails.
 */
sfsistat mlfi_abort(SMFICTX *ctx)
{
    /* Call clean-up function, nothing more to be done */
	return mlfi_cleanup(ctx, TRUE);
}

/* mlfi_cleanup:
 Called internally at end of processing to free memory used for this context
 */
sfsistat mlfi_cleanup(SMFICTX *ctx, bool okay)
{
    /* Grab our private session data */
    mlfiPriv_t *priv = MLFIPRIV;
    if (priv == NULL) {
		/* This shouldn't ever happen, but to be safe we test for it */
		MemErrorPrint("mlfi_cleanup", NULL, 1);
		if (alwaysAccept)
			return SMFIS_CONTINUE;
		else
			return SMFIS_TEMPFAIL;
	}
	
	/* Free recipient list */
	toAddrList_t *ptrT = priv->mlfi_toAddr_list;
	while (ptrT != NULL) {
		priv->mlfi_toAddr_list = ptrT->next;
		free(ptrT->address);
		free(ptrT);
		ptrT = priv->mlfi_toAddr_list;
	}
    
    /* Free header list */
    headerList_t *ptrH = priv->mlfi_header_list;
    while (ptrH != NULL) {
        // Free the header struct
        priv->mlfi_header_list = ptrH->next;
        if (ptrH->headerName != NULL)
            free(ptrH->headerName);
        if (ptrH->headerValue != NULL)
            free(ptrH->headerValue);
        free(ptrH);
        ptrH = priv->mlfi_header_list;
    }
    
    /* Free body chunks */
    bodyChunks_t *ptrB = priv->mlfi_body_chunks;
    while (ptrB != NULL) {
        priv->mlfi_body_chunks = ptrB->next;
        free(ptrB->body);
        free(ptrB);
        ptrB = priv->mlfi_body_chunks;
    }
    
    /* Milter API docs say to set ptr to null before terminating */
    free(priv);
    smfi_setpriv(ctx, NULL);
    
	if (!okay && !alwaysAccept) {
		/* We failed to collect this mail, tell the mta to reject it with a
		 temporary failure message so they will try again. */
		return SMFIS_TEMPFAIL;
	}
	/* Let mta continue processing */
	return SMFIS_CONTINUE;
}

/* Filter Registration Data, Name, API version, call backs to register for */
struct smfiDesc smfilter =
{
	"ArchiveSMTP",	/* filter name */
	SMFI_VERSION,	/* version code -- do not change */
	SMFIF_ADDRCPT,  /* flags */
	NULL,           /* connection info filter */
	NULL,           /* SMTP HELO command filter */
	mlfi_envfrom,	/* envelope sender filter */
	mlfi_envrcpt,   /* envelope recipient filter */
	mlfi_header,	/* header filter */
	NULL,           /* end of header */
	mlfi_body,      /* body block filter */
	mlfi_eom,       /* end of message */
	mlfi_abort,     /* message aborted */
	NULL,           /* connection cleanup */
	NULL,           /* unknown SMTP commands */
	NULL,           /* DATA command */
	NULL            /* Once, at the start of each SMTP connection */
};

/* ParseMailAddr: Removes the leading '<' and trailing '>' from a mail address
 and returns a newly allocted string containing just the address. */
char* ParseMailAddr(const char *addr)
{
	char *addrcopy;
	int len = strlen(addr);
	int i, j;
	
	/* Check that the address actually starts with '<' and ends with '>' */
	if (addr[0] == '<' && addr[(len-1)] == '>') {
		addrcopy = malloc(sizeof(char)*(len+1 - 2));
		if (addrcopy == NULL) {
			return (NULL);
		}
		i = 1;
		len = (len-2);
		addrcopy[len] = '\0';
	} else {
		addrcopy = malloc(sizeof(char)*(len+1));
		if (addrcopy == NULL) {
			return (NULL);
		}
		i = 0;
		len = (len+1);
	}
	
	/* Copy characters into the new string, changing them all the lower case */
	for (j = 0; j < len; i++, j++) {
		addrcopy[j] = tolower(addr[i]);
	}
	
	return (addrcopy);
}

/* MemErrorPrint: Prints a message to stderr about a failed memory allocation */
void MemErrorPrint(const char *func, const char *obj, int type)
{
	char *action;
	char action1[] = "Mail accepted despite error";
	char action2[] = "Mail returned to sender";
	
	if (alwaysAccept)
		action = action1; 
	else
		action = action2;
	
	switch (type) {
		case 0:
			fprintf(stderr, "ArchiveSMTP(%s): ERROR - Could not allocate memory for %s.\
					perhaps system or process memory limits have been reached? %s.\n",
					func, obj, action);
			break;
		case 1:
			fprintf(stderr, "ArchiveSMTP(%s): ERROR - Private data was not found!\
					Memory has been corrupted? %s.\n", func, action);
			break;
		case 2:
			fprintf(stderr, "ArchiveSMTP(%s): ERROR - %s (errno: %d:%s). %s.\n", func, obj, errno, strerror(errno), action);
			break;
		default:
			break;
	}
}

/*
 EvaluateRule's purpose is mainly to hide the implementation detail
 of evaluating the rule's truth or not, even though that isn't super
 complicated.
 */
int EvaluateRule(Rule* r, mlfiPriv_t *priv)
{
	return EvaluateCondition(r->root, priv);
}

void InitialiseCondition(Condition* cond)
{
	cond->t1 = cond->t2 = NULL;
	cond->p1 = cond->p2 = NULL;
	cond->type = kConditionTypeUndefined;
}

/*
	PrintTest(Test*)
	
	Prints the contents of a test to standard output. Used for debugging
	one's config file.
*/
void PrintTest(Test* t)
{
	switch (t->fieldToTest)
	{
		case kFieldSMTPSender:
			printf("SMTP Sender ");
			break;
		case kFieldFromAddress:
			printf("From Address ");
			break;
		case kFieldToAddress:
			printf("To Address ");
			break;
		case kFieldUndefined:
			printf("[Undefined Field] ");
	}
	
	switch (t->locationToTest)
	{
		case kLocationFieldStart:
			printf("starts with ");
			break;
		case kLocationFieldEnd:
			printf("ends with ");
			break;
		case kLocationFieldAny:
			printf("contains ");
			break;
		case kLocationFieldUndefined:
			printf("is operated on in some way by ");
			break;
	}
	
	printf("\"%s\"", t->textToFind);
}

/*
	PrintCondition(Condition*)
	
	A recursive function that prints the contents of the pointed-to
	passed condition, including its children. Also makes use of
	PrintTest(Test*).
*/
void PrintCondition(Condition* c)
{
	printf("(");
	if (c->p1)
		PrintCondition(c->p1);
	else if (c->t1)
		PrintTest(c->t1);
	if (c->type == kConditionTypeOr)
		printf(" or ");
	else if (c->type == kConditionTypeAnd)
		printf(" and ");
	if (c->p2)
		PrintCondition(c->p2);
	else if (c->t2)
		PrintTest(c->t2);
	printf(")");
}

/* PerformTest: Run the test on the current data and return the result. */
int PerformTest(Test* t, const char* testData)
{
	int result = FALSE;
		
	// See what part of the data we are testing and test it
	if (testData)
	{
		VerbosePrint(1, "\"%s\" ", testData);
		// Check length, if test data is less than find text no need to test
		int lengthOfTextToFind = strlen(t->textToFind);
		int lengthOfStringToTest = strlen(testData);
		if (lengthOfStringToTest >= lengthOfTextToFind) {
			int i,j;
			switch (t->locationToTest) {
				case kLocationFieldStart:		// Find text at start of data
					VerbosePrint(1, "%s ", "starts");
					for (i = 0; i < lengthOfTextToFind; i++) {
						if (t->textToFind[i] != testData[i])
						{
							result = FALSE;
							break;
						} else {
							result = TRUE;
						}
					}
					break;
				case kLocationFieldEnd:			// Find text at end of data
					VerbosePrint(1, "%s ", "ends");
					for (i = lengthOfTextToFind-1, j = lengthOfStringToTest-1; i >= 0; i--, j--) {
						if (t->textToFind[i] != testData[j]) {
							result = FALSE;
							break;
						} else {
							result = TRUE;
						}
					}
					break;
				case kLocationFieldAny:			// Find text any where in data
					VerbosePrint(1, "%s ", "contains");
					if (strstr(testData, t->textToFind)) {
						result = TRUE;
					} else {
						result = FALSE;
					}
					break;
				case kLocationFieldUndefined:
				default:
					VerbosePrint(1, "%s ", "confused");
					break;
			}
		}
	} else if (verboseLevel > 0) {
		VerbosePrint(1, "%s ", "[no data]");
		switch (t->locationToTest) {
			case kLocationFieldStart: VerbosePrint(1, "%s ", "starts"); break;
			case kLocationFieldEnd: VerbosePrint(1, "%s ", "ends"); break;
			case kLocationFieldAny: VerbosePrint(1, "%s ", "contains"); break;
			default: VerbosePrint(1, "%s ", "confused"); break;
		}
	}
	VerbosePrint(1, "\"%s\"", t->textToFind)
	
	if (result) {
		VerbosePrint(1, "%s", "[true]");
	} else {
		VerbosePrint(1, "%s", "[false]");
	}
	
	return (result);
}

int EvaluateConditionParameter(Condition* c, Test* t, mlfiPriv_t *priv)
{
	if (t) /* do we have a test in slot? */
	{
		/* yes! */
		toAddrList_t *ptr = priv->mlfi_toAddr_list;
		switch (t->fieldToTest) {
			case kFieldSMTPSender:
				return PerformTest(t, priv->mlfi_username);
				break;
			case kFieldFromAddress:
				return PerformTest(t, priv->mlfi_useraddr);
				break;
			case kFieldToAddress:
				/* There can be multiple to addresses, so test all of them */
				while (ptr != NULL) {
					if (PerformTest(t, ptr->address)) 
						return (TRUE);		/* Return true on the first match */
					VerbosePrint(1, "%s ", ",");
					ptr = ptr->next;
				}
				/* If we didn't return by this point, return false */
			case kFieldUndefined:
			default:
				return (FALSE);
				break;
		}
	}
	else if (c) /* no, so do we have a condition in slot? */
	{
		/* yes! */
		return EvaluateCondition(c, priv);
	}
	else /* we had no condition or test, so return true. */
		return TRUE;
}

int EvaluateCondition(Condition* cond, mlfiPriv_t *priv)
{	
	int firstEval, secondEval;
	
	VerbosePrint(1, "%s", "(");
	
	firstEval = EvaluateConditionParameter(cond->p1, cond->t1, priv);
	if (cond->type == kConditionTypeUndefined)
	{
		VerbosePrint(1, "%s", ")");
		return firstEval;
	}
	else if (cond->type == kConditionTypeAnd)
	{
		if (!firstEval)
		{
			VerbosePrint(1, "%s", ")");
			return FALSE;
		}
	}
	else /* kConditionTypeOr */
	{
		if (firstEval)
		{
			VerbosePrint(1, "%s", ")");
			return TRUE;
		}
	}
	
	secondEval = EvaluateConditionParameter(cond->p2, cond->t2, priv);
	
	VerbosePrint(1, "%s", ")");

	return secondEval;
	/* this is sort of weird, but here's the deal. If it's an AND we already
		know the first branch was true because otherwise we would've short
		circuited out by now. If it's an OR then we already know that the
		first condition was false, because otherwise we'd have short circuited
		out by now. If it's an Undefined then it can't get here either because
		it always drops out after the first condition.
		
		So the end result is that it's either
		TRUE && secondEval    or
		FALSE || secondEval
		both of which are functionally == to secondEval. */
}

/* allocate a new rule. The rule is a pristine slate ready for writing on. */
Rule* CreateNewRule()
{
	Rule* r = malloc(sizeof(Rule));
	r->root = NULL;
	return r;
}

void DeleteCondition(Condition* c)
{
	if (c->t1)
		free(c->t1);
	else if (c->p1)
		DeleteCondition(c->p1);
	if (c->t2)
		free(c->t2);
	else if (c->p2)
		DeleteCondition(c->p2);
	free(c);
}

void DeleteRule(Rule* r)
{
	DeleteCondition(r->root);
}

/* EatWhiteSpace: Moves forward in a file, eating white space characters until it
 finds non-white space, returning the first non-white space character it finds. */
char EatWhitespace(FILE* file)
{
	char c;
	int inComment = 0;
	while (1)
	{
		switch (c = fgetc(file))
		{
			case ' ':
			case '\t':
			case '\v':
			case '\r':
				break;
			case '\n':
				/* End comment if in one */
				if (inComment)
					inComment = FALSE;
				break;
			case '#':
				/* Also eat comments */
				if (!inComment)
					inComment = TRUE;
				break;
			default:
				/* Return the character if its not part of a comment */
				if (!inComment)
					return c;
				break;
		}
	}
}

/* EatWhiteAndReadString: Moves forward in a file, eating white space characters
 until it finds non-white space. The characters preceeding the white space until
 the next white space are read into a buffer and returned as a newly allocted
 string. Quote characters and backslash escapes are honoured, the buffer is limited
 to 256 bytes. */
char* EatWhiteAndReadString(FILE* file)
{
	char	buffer[256];
	char	c;
	char	*theStr;
	int		parse = TRUE;
	int		quote = FALSE;
	int		i = 0;
	
	/* Eat any leading whitespace */
	c = EatWhitespace(file);
	/* Parse file for a string, including quotes and escapes */
	while(parse) {
		switch(c) {
			case '\"':				/* Open or Close quotes */
				if (!quote)
					quote = TRUE;
				else
					quote = FALSE;
				i--;
				break;
			case '\\':				/* Escape - Read next character literally */
				buffer[i] = fgetc(file);
				break;
			case '\0':				/* this is always an error */
				fprintf(stderr, "Unexpected end-of-file while trying to parse string.\n");
				exit(EX_SOFTWARE);
				break;
			case ')':				/* Close bracket */
			case '}':				/* Close rule */
				if (!quote)
					ungetc(c, file);
			case ' ':				/* White space */
			case '\t':
			case '\v':
			case '\r':
			case '\n':
				if (!quote) {
					parse = FALSE;
					buffer[i] = '\0';
					break;
				}
			default:				/* All other characters */
				buffer[i] = c;
				break;
		}
		/* Get next character if still parsing */
		if (parse) {
			c = fgetc(file);
			i++;
			
			/* Explode on buffer overflow */
			if (i >= 256) {
				fprintf(stderr, "Buffer overflow! Perhaps the configuration file is missing a closing quote?\n");
				exit(EX_SOFTWARE);
			}
		}
	}
	
	/* Allocate and return the string */
	theStr = malloc(sizeof(char)*(i + 1));
	strcpy(theStr, buffer);
	
	return (theStr);
}

/* StringToLower: Goes through a null terminated string, turning any uppercase
 characters to lowercase. The original string is modified. */
void StringToLower(char *str)
{
	int i;
	for (i = 0; i < (strlen(str)); i++) {
		str[i] = tolower(str[i]);
	}
}

Test* ParseTest(FILE* file, int testField)
{
	Test*	theTest;
	char	input;
	
	theTest = malloc(sizeof(Test));
	theTest->fieldToTest = testField;
	/*
		Currently, the file should be up to a whitespace after the field to test was identified.
		
		1. eat whitespace
		2. read in the next word, the location in the field
		3. eat more whitespace
		4. read in the text to check
	*/
	
	input = EatWhitespace(file);
	switch (input)
	{
		case 'S':
		case 's':
			/* start */
			CheckStringIsNextInFile(file, "tarts", "Parsing error; expected \"Starts\"\n");
			theTest->locationToTest = kLocationFieldStart;
			break;
		case 'E':
		case 'e':
			/* end */
			CheckStringIsNextInFile(file, "nds", "Parsing error; expected \"Ends\"\n");
			theTest->locationToTest = kLocationFieldEnd;
			break;
		case 'C':
		case 'c':
			/* anywhere */
			CheckStringIsNextInFile(file, "ontains","Parsing error; expected \"Contains\"\n");
			theTest->locationToTest = kLocationFieldAny;
			break;
		default:
			fprintf(stderr, "Parsing error; expected a field location\n");
			exit(EX_SOFTWARE);
			break;
	}
	
	/* next item should be the string to search for. We have to read it in. */
	theTest->textToFind = EatWhiteAndReadString(file);
	
	/* convert to all lower-case if its not a username */
	if (testField != kFieldSMTPSender)
		StringToLower(theTest->textToFind);
	
	return (theTest);
}

/*  Returns TRUE if the passed string is next in the file and eats it.
	Returns FALSE otherwise. */
int CheckStringIsNextInFile(FILE* file, const char* string, const char* errorMessage)
{
	int i;
	char c;
	for (i = 0; string[i]; i++)
	{
		c = tolower(fgetc(file));
		if (c != string[i])
		{
			fprintf(stderr, errorMessage);
			exit(EX_SOFTWARE);
		}
		/*	return FALSE; */
	}
	return TRUE;
}

Condition* ParseCondition(FILE* file)
{
	/*
		Parse a condition. We don't really know what the condition's deal is, just
		that it's a condition. When we've parsed it we return the condition.
		
		The first thing we should encounter is a Test or a bracket.
		
		If we find a bracket, we need to call ParseCondition() again from this
		point. 
	*/
	
	Condition	*cond;
	char		input;
	
	cond = malloc(sizeof(Condition));
	
	InitialiseCondition(cond);
	
	input = EatWhitespace(file);
	switch (input)
	{
		case ' ':
		case '\t':
		case '\v':
		case '\n':
		case '\r':
			/* whitespace - ignore */
			break;
		case '\0':
			/* unexpected EOF - panic */
			fprintf(stderr, "Parsing error - unexpected End of File\n");
			exit(EX_SOFTWARE);
		case 'S':
		case 's':
			/* S is hopefully the start of sender */
			CheckStringIsNextInFile(file, "ender", "Parsing error - expected Sender\n");
			cond->t1 = ParseTest(file, kFieldSMTPSender);
			break;
		case 'F':
		case 'f':
			/* F is hopefully "from" */
			CheckStringIsNextInFile(file, "rom", "Parsing error - expected From\n");
			cond->t1 = ParseTest(file, kFieldFromAddress);
			break;
		case 'T':
		case 't':
			/* T is hopefully "to" */
			CheckStringIsNextInFile(file, "o", "Parsing error - expected To\n");
			cond->t1 = ParseTest(file, kFieldToAddress);
			break;
		case '(':
			cond->p1 = ParseCondition(file);
			break;
		default:
			fprintf(stderr, "Parsing error - unexpected character %c\n", input);
			exit(EX_SOFTWARE);
			break;
	}
	
	input = EatWhitespace(file);
	
	switch (input)
	{
		case 'O':
		case 'o':
			/* OR hopefully */
			CheckStringIsNextInFile(file, "r", "Parsing error - expected OR\n");
			/* if we made it this far then it was OR, which means we have to start a new condition */
			cond->type = kConditionTypeOr;
			cond->p2 = ParseCondition(file);
			break;
		case 'a':
		case 'A':
			CheckStringIsNextInFile(file, "nd", "Parsing error - expected AND\n");
			cond->type = kConditionTypeAnd;
			cond->p2 = ParseCondition(file);
			break;
		case ')':
			/* we just got to the end of a bracket, so we can just return out. */
			return cond;
		case '}':
			/* this should be the end of the rule, so we return out */
			return cond;
		default:
			/* uncovered case, which means a fatal error */
			fprintf(stderr, "Parsing error - expected o, a or ), found \"%c\"\n", input);
			break;
	}
	return cond;
}

void RuleStartedParsingState(FILE* file, Rule* theRule)
{
	/*
			This parsing state is entered after we encounter an opening bracket,
		signifying that a rule definition has started.
		
		Here is a reference example of a rule:
		
		{	Sender equals Fred or From begins fred	} copy to "/Logs/"
	 */
	char	input;
	
	/* Read the details of the condition. This will stop when we reach the } */
	theRule->root = ParseCondition(file);
	/* so at this point we just found a closing brace. Let's eat the whitespace. */
	input = EatWhitespace(file);
	
	/* The next part should be the instruction for what to do with the mail. This
		is probably "store in" at this point, followed by a location. So check the
		first character; if it's not Ss, explode, and if it is then check for the
		rest of tore in. If it's not there then error out. */
	switch(input) {
		case 'S':
		case 's':
			CheckStringIsNextInFile(file, "tore in", "Parsing error - expected Store in\n");
			/* We now need to eat the whitespace, then read the rest of the filename that should
			 be here. */
			theRule->storeIn = EatWhiteAndReadString(file);
			break;
		default:
			/* Explode - No store in found after closing bracket */
			fprintf(stderr, "Parsing error - expected Store in\n");
			exit(EX_SOFTWARE);
			break;
	}
}

/* AddRuleToRuleList: Adds a new rule to our list of rules to process. */
void AddRuleToRuleList(Rule* newRule)
{
	// Iterate to end of rule list, last rule will be at p
	Rule *i = firstRule;
	Rule *p = NULL;
	while (i) {
		p = i;
		i = i->next;
	}
	
	// Add the new rule to the end of the rule list
	newRule->next = NULL;
	if (p)
		p->next = newRule;
	else
		firstRule = newRule;
}

void InitialParsingState(FILE* file)
{
	/*
		This is the first parsing state, called immediately after we start 
	 */
	char	input;
	int		parse = 1;
	int		inComment = 0;
	Rule*	newRule;
	while (parse)
	{
		input = fgetc(file);
		switch (input)
		{
			case '\0':
			case EOF:
				/* eof - stop parsing */
				parse = 0;
				break;
			case ' ':
			case '\t':
			case '\v':
			case '\r':
				/* whitespace - ignore */
				break;
			case '\n':
				/* terminate comment if one is open else ignore */
				if (inComment)
					inComment = FALSE;
				break;
			case '#':
				/* Start of a comment - ignore rest of line */
				if (!inComment)
					inComment = TRUE;
				break;
			case '{':
				/* open bracket - start a thingy */
				if (!inComment) {
					newRule = CreateNewRule();
					AddRuleToRuleList(newRule);
					RuleStartedParsingState(file, newRule);
				}
				break;
			default:
				/* found something else - panic unless in a comment */
				if (!inComment) {
					fprintf(stderr, "Parsing error - unexpected character: '%c'\n", input);
					exit(EX_SOFTWARE);
				}
				break;
		}
	}
}

void ReadConfigurationFile(const char* filename)
{
	FILE* file = fopen(filename, "r");
	
	if (!file)
	{
		fprintf(stderr, "Unable to open config file at \"%s\" (errno: %d:%s). Aborting.\n", filename, errno, strerror(errno));
		exit(EX_SOFTWARE);
	}
	
	/* file opened successfully, so we can start parsing. */
	InitialParsingState(file);
	
	fclose(file);
}

/* Usage string, should be used by 'getopt' magic */
static void usage(char *prog)
{
	fprintf(stderr,
            "Usage: %s [-adhnv] [-r pidfile] [-t timeout] [-u username] -f filename -p port\n",
            prog);
}

/* main:
 Archive SMTP Milter - Main (setup, register, pass off to libmilter main), code
 taken mostly as-is from milter example main.
 */
int main (int argc, char * const argv[])
{   
	/* Initialise Globals */
	firstRule = NULL;
	verboseLevel = 0;
	alwaysAccept = TRUE;
	
    /* Vars defined by getopt:
	 * extern char *optarg;
     * extern int optind;
     * extern int optopt;
     * extern int opterr;
     * extern int optreset;
	 */
	
    /* Our argument list */
    const char *args = "adf:hnp:r:t:u:v";
	
	/* Argument Set Flags */
	bool	sockPortSet = FALSE;
	bool	confFileSet = FALSE;
	bool	userNameSet = FALSE;
	bool	testConfig	= FALSE;
	bool	pidFileSet	= FALSE;
	bool	timeOutSet	= FALSE;
	bool	dangeresque = FALSE;
	
	/* Argument Strings */
	char	*sockPort	= NULL;
	char	*confFile	= NULL;
	char	*userName	= NULL;
	char	*pidFile	= NULL;
	char	*timeOut	= NULL;
    
	/* Process command line options */
	char c;
	while ((c = getopt(argc, argv, args)) != -1)
	{
		switch (c)
		{
			case 'a':	/* Always store mail, reject mail if storage fails */
				alwaysAccept = FALSE;
				break;
			case 'd':	/* Dangeresque mode, will unlink the socket as root */
				dangeresque = TRUE;
				break;
			case 'f':	/* Configuration file path		- REQUIRED */
				confFileSet = TRUE;
				confFile = optarg;
				break;
			case 'n':	/* Parse & Print config then exit */
				testConfig = TRUE;
				break;
			case 'p':	/* Port or Socket used for IPC	- REQUIRED */
				sockPortSet = TRUE;
				sockPort = optarg;
				break;
			case 'r':	/* File to store process id in */
				pidFileSet = TRUE;
				pidFile = optarg;
				break;
			case 't':	/* Timeout value on MTA comms */
				timeOutSet = TRUE;
				timeOut = optarg;
				break;
			case 'u':	/* Username to run as */
				userNameSet = TRUE;
				userName = optarg;
				break;
			case 'v':	/* Increase verbosity level */
				verboseLevel++;
				break;
			case 'h':	/* Just wants to see usage string */
			default:	/* Unkown argument, print usage string and exit */
                usage(argv[0]);
                exit(EX_USAGE);
				break;
		}
	}
	
	/* First action should be to switch user if one was given */
	if (userNameSet) {
		/* Check if they are root, only root can change user */
		if (getuid() != 0) {
			fprintf(stderr, "You must be root to use the: -u option\n");
			exit(EX_USAGE);
		}
		else if (userName == NULL || *userName == '\0') {
			usage(argv[0]);
			fprintf(stderr, "Missing argument to -u: username\n");
			exit(EX_USAGE);
		}
		
		/* Lookup username and attempt to switch uid/gid to that user */
		struct passwd *uid;
		uid = getpwnam(userName);
		if (uid == NULL) { 
			fprintf(stderr, "Unable to switch user to: %s - No such user\n", userName);
			exit(EX_USAGE);
		}
		else if (setgid(uid->pw_gid) != 0 || setuid(uid->pw_uid) != 0) {
			fprintf(stderr, "Unable to switch uid/gid to: %s (errno: %d:%s)\n", userName, errno, strerror(errno));
			exit(EX_SOFTWARE);
		}
	}
	
	/* Next we should parse the config file, this file must exist */
	if (confFileSet) {
		if (confFile == NULL || *confFile == '\0') {
			usage(argv[0]);
			fprintf(stderr, "Missing argument to -f: filename\n");
			exit(EX_USAGE);
		}
		
		/* Process configuration file */
		ReadConfigurationFile(confFile);
	} else {
		usage(argv[0]);
		fprintf(stderr, "Missing required argument: -f\n");
		exit(EX_USAGE);
	}
		
	/* If the testConfig option is set, we will just print the rules and exit */
	if (testConfig) {
		Rule *itr = firstRule;
		while (itr) {
			PrintCondition(itr->root);
			/* print the location of the output */
			printf(" to \"%s\"\n", itr->storeIn);
			itr = itr->next;
		}
		exit(EX_OK);
	}
	
	/* Finally we need to set the port or socket specified */
	if (sockPortSet) {
		if (sockPort == NULL || *sockPort == '\0') {
			usage(argv[0]);
			fprintf(stderr, "Missing argument to -p: port\n");
			exit(EX_USAGE);
		}

		/* Set the port/socket to use, it will be created by smfi_main() */
		if (smfi_setconn(sockPort) == MI_FAILURE) {
			fprintf(stderr, "Unable to set port to: %s\n", sockPort);
			exit(EX_SOFTWARE);
		}
		
		/* If we're using a local socket, make sure it doesn't already exist.
		 This will not be done when running as root unless the -d option has
		 been set. */
		struct stat sockPortStat;
		int	portType;
		if (!strncasecmp(sockPort, "unix:", 5)) {
			portType = 5;
		} else if (!strncasecmp(sockPort, "local:", 6)) {
			portType = 6;
		} else {
			portType = 0;
		}
		if (portType != 0) {
			if (stat(sockPort + portType, &sockPortStat) != -1) {
				if (geteuid() != 0 || dangeresque) {
					unlink(sockPort + portType);
				}
			} else if (errno != ENOENT) {
				fprintf(stderr, "Unable to set port to: %s - invalid path (errno: %d:%s)\n", sockPort, errno, strerror(errno));
			}
		}
	} else {
		usage(argv[0]);
		fprintf(stderr, "Missing required argument: -p\n");
		exit(EX_USAGE);
	}
	
	/* Now that we are up and running, write our pid to file if set */
	int pidFD = 0;
	struct stat pidStat;
	if (pidFileSet) {
		if (pidFile == NULL || *pidFile == '\0') {
			usage(argv[0]);
			fprintf(stderr, "Missing argument to -r: pidfile\n");
			exit(EX_USAGE);
		}
		/* Try to open the pid file or create if it doesn't exist */
		pidFD = open(pidFile, O_RDWR | O_CREAT, 0600);
		if (pidFD == -1) {
			fprintf(stderr, "Unable to write pid to: %s (errno: %d:%s)\n", pidFile, errno, strerror(errno));
			exit(EX_SOFTWARE);
		}
		/* Try to obtain exclusive lock on the whole file */
		if (flock(pidFD, LOCK_EX | LOCK_NB) == -1) {
			if (errno == EWOULDBLOCK) {
				/* File is already locked by another instance of the program */
				char buffer[16];
				int i = read(pidFD, buffer, (sizeof(buffer) - 1));
				if (i == -1) {
					fprintf(stderr, "Program is already running, but pidfile was unreadable (errno: %d:%s)\n", errno, strerror(errno));
				} else {
					buffer[i] = '\0';
					fprintf(stderr, "Program is already running, pid: %s\n", buffer);
				}
			} else {
				fprintf(stderr, "Error, pidfile \"%s\" is not a regular file (errno: %d:%s)\n", pidFile, errno, strerror(errno));
			}
			exit(EX_SOFTWARE);
		}
		/* Record pid file stats and write pid to file */
		if (!ftruncate(pidFD, 0)) {
			char pidstr[16];
			snprintf(pidstr, sizeof(pidstr), "%u", getpid());
			if (pwrite(pidFD, pidstr, strlen(pidstr), 0) != strlen(pidstr)) {
				fprintf(stderr, "Unable to write pid to: %s (errno: %d:%s)\n", pidFile, errno, strerror(errno));
				exit(EX_SOFTWARE);
			}
		} else {
			fprintf(stderr, "Unable to truncate pidfile: %s (errno: %d:%s)\n", pidFile, errno, strerror(errno));
			exit(EX_SOFTWARE);
		}
		if (fstat(pidFD, &pidStat)) {
			fprintf(stderr, "Unable to fstat pid file: %s (errno: %d:%s)\n", pidFile, errno, strerror(errno));
			exit(EX_SOFTWARE);
		}
	}
	
	/* Configure the timeout for MTA read/writes if one was given */
	if (timeOutSet) {
		if (timeOut == NULL || *timeOut == '\0') {
			usage(argv[0]);
			fprintf(stderr, "Missing argument to -t: timeout\n");
			exit(EX_USAGE);
		}
		
		/* Set the time out */
		if (smfi_settimeout(atoi(timeOut)) == MI_FAILURE)
		{
			fprintf(stderr, "Unable to set timeout to: %s\n", timeOut);
			exit(EX_SOFTWARE);
		}
	}

    /* Register fiter with MTA */
	if (smfi_register(smfilter) == MI_FAILURE)
	{
		fprintf(stderr, "Unable to register with MTA\n");
		exit(EX_UNAVAILABLE);
	}
    
    /* Pass off to milters main */
    int exitVal = smfi_main();
	int exitErr = errno;
    
    /* Clean up globals before exit */
	Rule *ptr = firstRule, *oldPtr;
	while (ptr) {
		// Delete stuff
		DeleteRule(ptr);
		oldPtr = ptr;
		ptr = ptr->next;
		free(oldPtr);
	}
	
	if (pidFileSet) {
		/* Delete pid file if stats match saved values */
		struct stat pidFileStat;
		stat(pidFile, &pidFileStat);
		if (pidFileStat.st_dev == pidStat.st_dev &&		/* Check device id */
			pidFileStat.st_ino == pidStat.st_ino &&		/* Check inode */
			pidFileStat.st_mode == pidStat.st_mode &&	/* Check access modes */
			pidFileStat.st_uid == pidStat.st_uid &&		/* Check owner */
			pidFileStat.st_gid == pidStat.st_gid &&		/* Check group */
			pidFileStat.st_rdev == pidStat.st_rdev) {	/* Check the file type */
			/* Perhaps all these checks are over the top? There could still be race condition..? */
			if (unlink(pidFile) == -1) {
				fprintf(stderr, "Unable to remove pidfile \"%s\" on exit (errno: %d:%s)\n", pidFile, errno, strerror(errno));
			}
		} else {
			fprintf(stderr, "Aborting removal of pidfile \"%s\" on exit, file has been changed\n", pidFile);
		}
		if (flock(pidFD, LOCK_UN) == -1) {
			fprintf(stderr, "Unable to remove lock on pidfile \"%s\" on exit (errno: %d:%s)\n", pidFile, errno, strerror(errno));
		}
		if (close(pidFD) == -1) {
			fprintf(stderr, "Unable to close pidfile \"%s\" on exit (errno: %d:%s)\n", pidFile, errno, strerror(errno));
		}
	}
    
	if (exitVal != 0)
		fprintf(stderr, "Exiting with error, this usually means there was a problem with the socket (errno: %d:%s)\n", exitErr, strerror(exitErr));
	
    return (exitVal);
}
